__all__ = ["__Placeholder"]


class __Placeholder:
    def __add__(self, _):
        return self

    def __sub__(self, _):
        return self

    def __mul__(self, _):
        return self

    def __truediv__(self, _):
        return self

    def __floordiv__(self, _):
        return self

    def __mod__(self, _):
        return self

    def __pow__(self, _):
        return self

    def __radd__(self, _):
        return self

    def __rsub__(self, _):
        return self

    def __rmul__(self, _):
        return self

    def __rtruediv__(self, _):
        return self

    def __rfloordiv__(self, _):
        return self

    def __rmod__(self, _):
        return self

    def __rpow__(self, _):
        return self

    def __neg__(self):
        return self

    def __pos__(self):
        return self

    def __not__(self):
        return self

    def __abs__(self):
        return self

    def __invert__(self):
        return self

    def __eq__(self, _):
        return False

    def __ne__(self, _):
        return True

    def __lt__(self, _):
        return False

    def __le__(self, _):
        return False

    def __gt__(self, _):
        return False

    def __ge__(self, _):
        return False

    def __bool__(self):
        return False

    def __call__(self, *_, **__):
        return self

    def __getitem__(self, _):
        return self

    def __repr__(self):
        return "__Placeholder()"

    def __str__(self):
        return "Placeholder"
